@@ -75,6 +75,7 @@ struct _CcApplicationsPanel
   GSettings       *search_settings;
 
   GtkListBox      *stack;
+  GtkButton       *install_button;
 
   GtkWidget       *permission_section;
   GtkWidget       *permission_list;
@@ -140,6 +141,20 @@ enum
   PROP_PARAMETERS
 };
 
+static gboolean
+gnome_software_is_installed (void)
+{
+  g_autofree gchar *path = g_find_program_in_path ("gnome-software");
+  return path != NULL;
+}
+
+static gboolean
+snap_store_is_installed (void)
+{
+  g_autofree gchar *path = g_find_program_in_path ("snap-store");
+  return path != NULL;
+}
+
 /* Callbacks */
 
 static gboolean
@@ -160,6 +175,9 @@ open_software_cb (GtkButton           *button,
 {
   const gchar *argv[] = { "gnome-software", "--details", "appid", NULL };
 
+  if (!gnome_software_is_installed ())
+    argv[0] = "snap-store";
+
   if (self->current_app_id == NULL)
     argv[1] = NULL;
   else
@@ -630,10 +648,13 @@ add_snap_permissions (CcApplicationsPanel *self,
   g_autoptr(GList) rows = NULL;
   gint index;
   g_autoptr(SnapdClient) client = NULL;
+  g_autoptr(GPtrArray) interfaces = NULL;
   g_autoptr(GPtrArray) plugs = NULL;
   g_autoptr(GPtrArray) slots = NULL;
+  SnapdInterface *interface = NULL;
   gint added = 0;
   g_autoptr(GError) error = NULL;
+  g_autoptr(GError) interfaces_error = NULL;
 
   if (!g_str_has_prefix (app_id, PORTAL_SNAP_PREFIX))
     return FALSE;
@@ -644,14 +665,22 @@ add_snap_permissions (CcApplicationsPanel *self,
   g_assert (index >= 0);
 
   client = snapd_client_new ();
+
+  interfaces = snapd_client_get_interfaces2_sync (client,
+                                                  SNAPD_GET_INTERFACES_FLAGS_NONE,
+                                                  NULL,
+                                                  NULL, &interfaces_error);
+  if (interfaces == NULL)
+    g_warning ("Failed to get snap interfaces: %s", interfaces_error->message);
+
   if (!snapd_client_get_connections2_sync (client,
-                                           SNAPD_GET_CONNECTIONS_FLAGS_NONE,
-                                           snap_name, NULL,
+                                           SNAPD_GET_CONNECTIONS_FLAGS_SELECT_ALL,
+                                           NULL, NULL,
                                            NULL, NULL,
                                            &plugs, &slots,
                                            NULL, &error))
     {
-      g_warning ("Failed to get snap interfaces: %s", error->message);
+      g_warning ("Failed to get snap connections: %s", error->message);
       return FALSE;
     }
 
@@ -668,6 +697,10 @@ add_snap_permissions (CcApplicationsPanel *self,
                                                   "x11",
                                                   NULL };
 
+      /* Skip if not relating to this snap */
+      if (g_strcmp0 (snapd_plug_get_snap (plug), snap_name) != 0)
+        continue;
+
       /* Ignore interfaces that are too low level to make sense to show or disable */
       if (g_strv_contains (hidden_interfaces, snapd_plug_get_interface (plug)))
         continue;
@@ -682,7 +715,17 @@ add_snap_permissions (CcApplicationsPanel *self,
           g_ptr_array_add (available_slots, g_object_ref (slot));
         }
 
-      row = cc_snap_row_new (cc_panel_get_cancellable (CC_PANEL (self)), plug, available_slots);
+      if (interfaces != NULL)
+        {
+          for (int j = 0; j < interfaces->len; j++)
+            {
+              SnapdInterface *i = g_ptr_array_index (interfaces, j);
+              if (g_strcmp0 (snapd_interface_get_name (i), snapd_plug_get_interface (plug)) == 0)
+                interface = i;
+            }
+        }
+
+      row = cc_snap_row_new (cc_panel_get_cancellable (CC_PANEL (self)), interface, plug, available_slots);
       gtk_widget_show (GTK_WIDGET (row));
       gtk_list_box_insert (GTK_LIST_BOX (self->permission_list), GTK_WIDGET (row), index);
       index++;
@@ -1582,7 +1625,7 @@ update_panel (CcApplicationsPanel *self,
 
   gtk_label_set_label (GTK_LABEL (self->title_label), g_app_info_get_display_name (info));
   gtk_stack_set_visible_child_name (GTK_STACK (self->stack), "settings");
-  gtk_widget_show (self->header_button);
+  gtk_widget_set_visible (self->header_button, gnome_software_is_installed () || snap_store_is_installed ());
 
   g_clear_pointer (&self->current_app_id, g_free);
   g_clear_pointer (&self->current_portal_app_id, g_free);
@@ -1872,6 +1915,7 @@ cc_applications_panel_class_init (CcApplicationsPanelClass *klass)
   gtk_widget_class_bind_template_child (widget_class, CcApplicationsPanel, handler_section);
   gtk_widget_class_bind_template_child (widget_class, CcApplicationsPanel, handler_reset);
   gtk_widget_class_bind_template_child (widget_class, CcApplicationsPanel, handler_list);
+  gtk_widget_class_bind_template_child (widget_class, CcApplicationsPanel, install_button);
   gtk_widget_class_bind_template_child (widget_class, CcApplicationsPanel, integration_list);
   gtk_widget_class_bind_template_child (widget_class, CcApplicationsPanel, integration_section);
   gtk_widget_class_bind_template_child (widget_class, CcApplicationsPanel, location);
@@ -1930,6 +1974,8 @@ cc_applications_panel_init (CcApplicationsPanel *self)
 
   gtk_widget_init_template (GTK_WIDGET (self));
 
+  gtk_widget_set_visible (GTK_WIDGET (self->install_button), gnome_software_is_installed () || snap_store_is_installed ());
+
   provider = GTK_STYLE_PROVIDER (gtk_css_provider_new ());
   gtk_css_provider_load_from_resource (GTK_CSS_PROVIDER (provider),
                                        "/org/gnome/control-center/applications/cc-applications-panel.css");
