@@ -29,10 +29,6 @@
 
 #define MAX_SCALE_BUTTONS 5
 
-#define MUTTER_SCHEMA                     "org.gnome.mutter"
-#define MUTTER_EXPERIMENTAL_FEATURES_KEY  "experimental-features"
-#define MUTTER_FEATURE_FRACTIONAL_SCALING "x11-randr-fractional-scaling"
-
 struct _CcDisplaySettings
 {
   GtkDrawingArea    object;
@@ -43,7 +39,6 @@ struct _CcDisplaySettings
   gboolean          has_accelerometer;
   CcDisplayConfig  *config;
   CcDisplayMonitor *selected_output;
-  GSettings        *mutter_settings;
 
   GListStore       *orientation_list;
   GListStore       *refresh_rate_list;
@@ -78,49 +73,6 @@ static void on_scale_btn_active_changed_cb (GtkWidget         *widget,
                                             GParamSpec        *pspec,
                                             CcDisplaySettings *self);
 
-
-static gboolean
-get_fractional_scaling_active (GValue *value, GVariant *variant, gpointer user_data)
-{
-  g_autofree const gchar **features = NULL;
-
-  features = g_variant_get_strv (variant, NULL);
-  g_value_set_boolean (value, g_strv_contains (features, MUTTER_FEATURE_FRACTIONAL_SCALING));
-  return TRUE;
-}
-
-static GVariant *
-set_fractional_scaling_active (const GValue *value, const GVariantType *expected_type, gpointer user_data)
-{
-  CcDisplaySettings *self = user_data;
-  gboolean enable;
-  g_auto(GStrv) existing_features = NULL;
-  gboolean have_fractional_scaling = FALSE;
-  g_autoptr(GVariantBuilder) builder = NULL;
-
-  enable = g_value_get_boolean (value);
-
-  /* Add or remove the fractional scaling feature from mutter */
-  existing_features = g_settings_get_strv (self->mutter_settings, MUTTER_EXPERIMENTAL_FEATURES_KEY);
-  builder = g_variant_builder_new (G_VARIANT_TYPE ("as"));
-  for (int i = 0; existing_features[i] != NULL; i++)
-    {
-      if (strcmp (existing_features[i], MUTTER_FEATURE_FRACTIONAL_SCALING) == 0)
-        {
-          if (enable)
-            have_fractional_scaling = TRUE;
-          else
-            continue;
-        }
-
-      g_variant_builder_add (builder, "s", existing_features[i]);
-    }
-  if (enable && !have_fractional_scaling)
-    g_variant_builder_add (builder, "s", MUTTER_FEATURE_FRACTIONAL_SCALING);
-
-  return g_variant_builder_end (builder);
-}
-
 static gboolean
 should_show_rotation (CcDisplaySettings *self)
 {
@@ -480,6 +432,8 @@ cc_display_settings_rebuild_ui (CcDisplaySettings *self)
   gtk_widget_set_visible (self->scale_row, buttons > 1);
 
   gtk_widget_set_visible (self->scale_fractional_row, TRUE);
+  gtk_switch_set_active (GTK_SWITCH (self->scale_fractional_switch),
+                         cc_display_config_get_fractional_scaling (self->config));
 
   gtk_widget_set_visible (self->underscanning_row,
                           cc_display_monitor_supports_underscanning (self->selected_output) &&
@@ -666,7 +620,6 @@ cc_display_settings_finalize (GObject *object)
   CcDisplaySettings *self = CC_DISPLAY_SETTINGS (object);
 
   g_clear_object (&self->config);
-  g_clear_object (&self->mutter_settings);
 
   g_clear_object (&self->orientation_list);
   g_clear_object (&self->refresh_rate_list);
@@ -735,19 +688,28 @@ cc_display_settings_class_init (CcDisplaySettingsClass *klass)
   gtk_widget_class_bind_template_callback (widget_class, on_underscanning_switch_active_changed_cb);
 }
 
+static void
+on_scale_fractional_toggled (CcDisplaySettings *self)
+{
+  gboolean active;
+
+  active = gtk_switch_get_active (GTK_SWITCH (self->scale_fractional_switch));
+
+  if (self->config)
+    cc_display_config_set_fractional_scaling (self->config, active);
+
+  g_signal_emit_by_name (G_OBJECT (self), "updated", self->selected_output);
+}
+
 static void
 cc_display_settings_init (CcDisplaySettings *self)
 {
   gtk_widget_init_template (GTK_WIDGET (self));
 
-  self->mutter_settings = g_settings_new (MUTTER_SCHEMA);
-  g_settings_bind_with_mapping (self->mutter_settings, MUTTER_EXPERIMENTAL_FEATURES_KEY,
-                                self->scale_fractional_switch, "active",
-                                G_SETTINGS_BIND_DEFAULT,
-                                get_fractional_scaling_active,
-                                set_fractional_scaling_active,
-                                self,
-                                NULL);
+  g_signal_connect_object (self->scale_fractional_switch,
+                           "notify::active",
+                           G_CALLBACK (on_scale_fractional_toggled),
+                           self, G_CONNECT_SWAPPED);
 
   gtk_list_box_set_header_func (GTK_LIST_BOX (self),
                                 cc_list_box_update_header_func,
