@@ -48,7 +48,8 @@
 #include "cc-realm-manager.h"
 #include "cc-user-accounts-resources.h"
 #include "cc-user-image.h"
-#include "um-fingerprint-dialog.h"
+#include "cc-fingerprint-manager.h"
+#include "cc-fingerprint-dialog.h"
 #include "user-utils.h"
 
 #include "cc-common-language.h"
@@ -103,6 +104,8 @@ struct _CcUserPanel {
 
         CcAvatarChooser *avatar_chooser;
 
+        CcFingerprintManager *fingerprint_manager;
+
         gint other_accounts;
 };
 
@@ -174,10 +177,9 @@ set_selected_user (CcUserPanel *self, CcCarouselItem *item)
 {
         uid_t uid;
 
-        g_clear_object (&self->selected_user);
-
         uid = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (item), "uid"));
-        self->selected_user = act_user_manager_get_user_by_id (self->um, uid);
+        g_set_object (&self->selected_user,
+                      act_user_manager_get_user_by_id (self->um, uid));
 
         if (self->selected_user != NULL) {
                 show_user (self->selected_user, self);
@@ -795,6 +797,25 @@ get_autologin_possible (ActUser *user)
 static void on_permission_changed (CcUserPanel *self);
 static void full_name_edit_button_toggled (CcUserPanel *self);
 
+static void
+update_fingerprint_row_state (CcUserPanel *self, GParamSpec *spec, CcFingerprintManager *fingerprint_manager)
+{
+        CcFingerprintState state = cc_fingerprint_manager_get_state (fingerprint_manager);
+
+        if (state != CC_FINGERPRINT_STATE_UPDATING) {
+                gtk_widget_set_visible (GTK_WIDGET (self->fingerprint_row),
+                                        state != CC_FINGERPRINT_STATE_NONE);
+        }
+
+        gtk_widget_set_sensitive (GTK_WIDGET (self->fingerprint_row),
+                                  state != CC_FINGERPRINT_STATE_UPDATING);
+
+        if (state == CC_FINGERPRINT_STATE_ENABLED)
+                gtk_label_set_text (self->fingerprint_state_label, _("Enabled"));
+        else if (state == CC_FINGERPRINT_STATE_DISABLED)
+                gtk_label_set_text (self->fingerprint_state_label, _("Disabled"));
+}
+
 static void
 show_user (ActUser *user, CcUserPanel *self)
 {
@@ -852,9 +873,22 @@ show_user (ActUser *user, CcUserPanel *self)
         show = (act_user_get_uid (user) == getuid() &&
                 act_user_is_local_account (user) &&
                 (self->login_screen_settings &&
-                 g_settings_get_boolean (self->login_screen_settings, "enable-fingerprint-authentication")) &&
-                set_fingerprint_label (self->fingerprint_state_label));
-        gtk_widget_set_visible (GTK_WIDGET (self->fingerprint_row), show);
+                 g_settings_get_boolean (self->login_screen_settings,
+                                         "enable-fingerprint-authentication")));
+
+        if (show) {
+                if (!self->fingerprint_manager) {
+                        self->fingerprint_manager = cc_fingerprint_manager_new (user);
+                        g_signal_connect_object (self->fingerprint_manager,
+                                                 "notify::state",
+                                                 G_CALLBACK (update_fingerprint_row_state),
+                                                 self, G_CONNECT_SWAPPED);
+                }
+
+                update_fingerprint_row_state (self, NULL, self->fingerprint_manager);
+        } else {
+                gtk_widget_set_visible (GTK_WIDGET (self->fingerprint_row), FALSE);
+        }
 
         /* Autologin: show when local account */
         show = act_user_is_local_account (user);
@@ -1084,12 +1118,17 @@ static void
 change_fingerprint (CcUserPanel *self)
 {
         ActUser *user;
+        GtkWindow *top_level;
+        CcFingerprintDialog *dialog;
 
         user = get_selected_user (self);
+        top_level = GTK_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (self)));
 
         g_assert (g_strcmp0 (g_get_user_name (), act_user_get_user_name (user)) == 0);
 
-        fingerprint_button_clicked (GTK_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (self))), self->fingerprint_state_label, user);
+        dialog = cc_fingerprint_dialog_new (self->fingerprint_manager);
+        gtk_window_set_transient_for (GTK_WINDOW (dialog), top_level);
+        gtk_widget_show (GTK_WIDGET (dialog));
 }
 
 static void
@@ -1235,16 +1274,9 @@ on_permission_changed (CcUserPanel *self)
         gboolean self_selected;
         ActUser *user;
 
-        user = get_selected_user (self);
-        if (!user) {
-                return;
-        }
-
         is_authorized = g_permission_get_allowed (G_PERMISSION (self->permission));
-        self_selected = act_user_get_uid (user) == geteuid ();
 
         gtk_widget_set_visible (GTK_WIDGET (self->add_user_button), is_authorized);
-
         gtk_widget_set_sensitive (GTK_WIDGET (self->add_user_button), is_authorized);
         if (is_authorized) {
                 setup_tooltip_with_embedded_icon (GTK_WIDGET (self->add_user_button), _("Create a user account"), NULL, NULL);
@@ -1264,6 +1296,12 @@ on_permission_changed (CcUserPanel *self)
                 g_object_unref (icon);
         }
 
+        user = get_selected_user (self);
+        if (!user) {
+                return;
+        }
+
+        self_selected = act_user_get_uid (user) == geteuid ();
         gtk_widget_set_sensitive (GTK_WIDGET (self->remove_user_button), is_authorized && !self_selected
                                   && !would_demote_only_admin (user));
         if (is_authorized) {
@@ -1328,6 +1366,11 @@ on_permission_changed (CcUserPanel *self)
         }
 
         if (is_authorized || self_selected) {
+                CcFingerprintState fingerprint_state = CC_FINGERPRINT_STATE_NONE;
+
+                if (self->fingerprint_manager)
+                        fingerprint_state = cc_fingerprint_manager_get_state (self->fingerprint_manager);
+
                 gtk_stack_set_visible_child (self->user_icon_stack, GTK_WIDGET (self->user_icon_button));
 
                 gtk_widget_set_sensitive (GTK_WIDGET (self->language_row), TRUE);
@@ -1336,7 +1379,8 @@ on_permission_changed (CcUserPanel *self)
                 gtk_widget_set_sensitive (GTK_WIDGET (self->password_row), TRUE);
                 remove_unlock_tooltip (GTK_WIDGET (self->password_row));
 
-                gtk_widget_set_sensitive (GTK_WIDGET (self->fingerprint_row), TRUE);
+                gtk_widget_set_sensitive (GTK_WIDGET (self->fingerprint_row),
+                                          fingerprint_state != CC_FINGERPRINT_STATE_UPDATING);
                 remove_unlock_tooltip (GTK_WIDGET (self->fingerprint_row));
 
                 gtk_widget_set_sensitive (GTK_WIDGET (self->last_login_row), TRUE);
@@ -1481,6 +1525,8 @@ cc_user_panel_dispose (GObject *object)
 {
         CcUserPanel *self = CC_USER_PANEL (object);
 
+        g_clear_object (&self->selected_user);
+
         g_clear_object (&self->login_screen_settings);
 
         g_clear_pointer ((GtkWidget **)&self->language_chooser, gtk_widget_destroy);
